#include <stdio.h>
#import "CPU.h"
#import "Bus.h"
#import "CPU_KbdInterface.h"
#import "Display_CPUInterface.h"
#import "DebugController.h"
#import "DataOperations.h"
#import "SoundOutput.h"

static float simulatedInstructionTime = 155e-6;
static float maxSimulationTimeLag = 0.2;

// Funky mapping from keyboard to CPU column numbers
static int keyColTable[] = {0x10, 0x30, 0x70, 0x80, 0xC0};

enum {
    onKeyCode = 0x18
};

#define A self->reg.A
#define B self->reg.B
#define C self->reg.C
#define M self->reg.M
#define N self->reg.N
#define G self->reg.G
#define ST self->reg.ST
#define XST self->reg.XST
#define T self->reg.T
#define FI self->reg.FI
#define PC self->reg.PC

#define P self->reg.PQ[0]
#define Q self->reg.PQ[1]
#define R self->reg.PQ[self->reg.r]
#define MODE self->reg.mode
#define UNIMP [self unimplementedInstruction]

static void DecrementPointer(Digit *p) {
    int x = *p - 1;
    if (x < 0)
        x = 13;
    *p = x;
}

static BOOL Fetch(CPU *self, int *data) {
    --self->cycleLimit;
    self->simulationTime += simulatedInstructionTime;
    SoundOutputForWordTime(T);
    return [self->bus readRomLocation: PC++ into: data];
}

static void PushReturnStack(CPU *self, int word) {
    int i;
    for (i = 3; i >= 1; i--)
        self->reg.stack[i] = self->reg.stack[i-1];
    self->reg.stack[0] = word;
}

static int PopReturnStack(CPU *self) {
    int i;
    int result = self->reg.stack[0];
    for (i = 0; i <= 3; i++)
        self->reg.stack[i] = self->reg.stack[i+1];
    return result;
}

//---------------------------- Class 0 ---------------------------------------------

static void Execute_Class0_Line0(CPU *self, int f) {
    switch (f) {
        case 0xE:
        case 0xF:
            UNIMP;
            break;
        default: // NOP
            ;
    }
}

static void Execute_Class0_Line1(CPU *self, int f) {
    switch (f) {
        case 0xE:
            UNIMP;
            break;
        case 0xF: // 3C4 ST=0
            ST = 0;
            break;
        default: // CLRF f
            if (f <= 7)
                ST &= ~(1 << f);
            else
                XST &= ~(1 << (f - 8));
    }
}

static void Execute_Class0_Line2(CPU *self, int f) {
    switch (f) {
        case 0xE:
            UNIMP;
            break;
        case 0xF: // 3C8 CLRKEY
            if (!self->reg.keyDown) {
                if (self->keyReleaseDelay)   // See comment in reset:
                    --self->keyReleaseDelay;
                else
                    self->reg.KY = 0;
            }
            break;
        default: // SETF f
            if (f <= 7)
                ST |= (1 << f);
            else
                XST |= (1 << (f - 8));
    }
}

static void Execute_Class0_Line3(CPU *self, int f) {
    switch (f) {
        case 0xE:
            UNIMP;
            break;
        case 0xF: // 3CC ?KEY
            self->nextCarry = self->reg.KY != 0;
            break;
        default: // ?FSET f
            if (f <= 7)
                self->nextCarry = (ST & (1 << f)) != 0;
            else
                self->nextCarry = (XST & (1 << (f - 8))) != 0;
    }
}

static void Execute_Class0_Line5(CPU *self, int f) {
    switch (f) {
        case 0xE:
            UNIMP;
            break;
        case 0xF: // 3DC R=R-1
            R = (R == 0) ? 13 : R - 1;
            break;
        default: // ?R= f
            self->nextCarry = (R == f);
    }
}

static void Execute_Class0_Line6(CPU *self, int param) {
    int temp;
    switch (param) {
        case 0x1: // 058 G=C
            DigitsToBitsWrap(C, &G, R, 2);
            break;
        case 0x2: // 098 C=G
            BitsToDigitsWrap(G, C, R, 2);
            break;
        case 0x5: // 158 M=C
            CopyDigits(C, M, 14);
            break;
        case 0x6: // 198 C=M
            CopyDigits(M, C, 14);
            break;
        case 0x7: // 1D8 C<>M
            ExchangeDigits(C, M, 0, 14);
            break;
        case 0x9: // 258 T=ST
            T = ST;
            break;
        case 0xB: // 2D8 ST<>T
            temp = ST;
            ST = T;
            T = temp;
            break;
        case 0xD: // 358 ST=C
            DigitsToBits(C, &ST, 8);
            break;
        case 0xE: // 398 C=ST
            BitsToDigits(ST, C, 2);
            break;
        case 0xF: // 3D8 C<>ST
            temp = ST;
            DigitsToBits(C, &ST, 8);
            BitsToDigits(temp, C, 2);
            break;
        default:
            UNIMP;
    }
}

static void Execute_Class0_Line7(CPU *self, int f) {
    switch (f) {
        case 0xE:
            UNIMP;
            break;
        case 0xF: // 3DC R=R+1
            R = (R >= 13) ? 0 : R + 1;
            break;
        default: // R= f
            R = f;
    }
}

static void Execute_Class0_Line8(CPU *self, int param) {
    switch (param) {
        case 0x0: // 020 XQ>GO
            PopReturnStack(self);
            break;
        case 0x1: // 060 POWOFF
            if (self->powerOffFlag) {
                //printf("POWOFF: going into deep sleep and resetting powerOffFlag\n");
                [self setPowerMode: deepSleep];
                self->powerOffFlag = NO;
            }
            else {
                //printf("POWOFF: going into light sleep\n");
                [self setPowerMode: lightSleep];
            }
            PC = 0;
            break;
        case 0x2: // 0A0 SLCT P
            self->reg.r = 0;
            break;
        case 0x3: // 0E0 SLCT Q
            self->reg.r = 1;
            break;
        case 0x4: // 120 ?P=Q
            self->nextCarry = (P == Q);
            break;
        case 0x5: // 160 ?LOWBAT
            // Our batteries NEVER go flat! Or your money back!
            break;
        case 0x6: // 1A0 A=B=C=0
            FillDigits(A, 0, 14);
            FillDigits(B, 0, 14);
            FillDigits(C, 0, 14);
            break;
        case 0x7: // 1E0 GOTO ADR
            DigitsToBits(&C[3], &PC, 16);
            break;
        case 0x8: // 220 C=KEY
            BitsToDigits(self->reg.KY, &C[3], 2);
            break;
        case 0x9: // 260 SETHEX
            self->reg.mode = HEX_MODE;
            break;
        case 0xA: // 2A0 SETDEC
            self->reg.mode = DEC_MODE;
            break;
        case 0xB: // 2E0 DSPOFF
            [self->display displayOff];
            break;
        case 0xC: // 320 DSPTOG
            [self->display displayToggle];
            break;
        case 0xD: // 360 C RTN
            if (self->reg.carry)
                PC = PopReturnStack(self);
            break;
        case 0xE: // 3A0 NC RTN
            if (!self->reg.carry)
                PC = PopReturnStack(self);
            break;
        case 0xF: // 3E0 RTN
            PC = PopReturnStack(self);
            break;
         default:
            UNIMP;
    }
}

static void Execute_Class0_LineB(CPU *self, int f) {
    switch (f) {
        case 0xE:
        case 0xF:
            UNIMP;
            break;
        default: // ?FI f
            self->nextCarry = (FI & (1 << f)) != 0;
    }
}

static void Execute_Class0_LineC(CPU *self, int param) {
    int addr, word;
    switch (param) {
        case 0x1: // 070 N=C
            CopyDigits(C, N, 14);
            break;
        case 0x2: // 0B0 C=N
            CopyDigits(N, C, 14);
            break;
        case 0x3: // 0F0 C<>N
            ExchangeDigits(C, N, 0, 14);
            break;
        case 0x4: // 130 LDI S&X
            Fetch(self, &word);
            BitsToDigits(word, C, 3);
            break;
        case 0x5: // 170 PUSH ADR
            DigitsToBits(&C[3], &word, 16);
            PushReturnStack(self, word);
            break;
        case 0x6: // 1B0 POP ADR
            word = PopReturnStack(self);
            BitsToDigits(word, &C[3], 4);
            break;
        case 0x9: // 270 RAM SLCT
            DigitsToBits(&C[1], &self->reg.ramchip, 6);
            self->reg.ramreg = C[0];
            break;
        case 0xB: // 2F0 WRITE DATA
            [self->bus writeRamLocation: self->reg.ramreg ofChip: self->reg.ramchip from: C];
            if (self->reg.peripheral) {
                //int i;
                //printf("Peripheral Write Data: %02X ", self->reg.peripheral);
                //for (i = 13; i >= 0; i--)
                //    printf("%X", C[i]);
                //printf("\n");
                [self->bus writeDataToPeripheral: self->reg.peripheral from: C];
            }
            break;
        case 0xC: // 330 FETCH S&X
            DigitsToBits(&C[3], &addr, 16);
            [self->bus readRomLocation: addr into: &word];
            BitsToDigits(word, C, 3);
            break;
        case 0xD: // 370 C=C OR A
            OrDigits(C, A, C, 0, 14);
            break;
        case 0xE: // 3B0 C=C AND A
            AndDigits(C, A, C, 0, 14);
            break;
        case 0xF: // 3F0 PRPH SLCT
            DigitsToBits(C, &self->reg.peripheral, 8);
            break;
        default:
            UNIMP;
    }
}

static void Execute_Class0_LineF(CPU *self, int param) {
    switch (param) {
        case 0x0:
        case 0xE:
        case 0xF:
            UNIMP;
            break;
        default: { // RCR f
            Digits_14 temp;
            CopyDigits(C, temp, param);
            CopyDigits(&C[param], C, 14 - param);
            CopyDigits(temp, &C[14 - param], param);
        }
    }
}

static int fTable[] = {
    /* 0 */ 0x3,
    /* 1 */ 0x4,
    /* 2 */ 0x5,
    /* 3 */ 0xA,
    /* 4 */ 0x8,
    /* 5 */ 0x6,
    /* 6 */ 0xB,
    /* 7 */ 0xE,
    /* 8 */ 0x2,
    /* 9 */ 0x9,
    /* A */ 0x7,
    /* B */ 0xD,
    /* C */ 0x1,
    /* D */ 0xC,
    /* E */ 0x0,
    /* F */ 0xF,
};

static void Execute_Class0(CPU *self, int instr) {
    int param = instr >> 6;
    int opcode = (instr >> 2) & 0xf;
    switch (opcode) {
        case 0x0:
            Execute_Class0_Line0(self, fTable[param]);
            break;
        case 0x1:
            Execute_Class0_Line1(self, fTable[param]);
            break;
        case 0x2:
            Execute_Class0_Line2(self, fTable[param]);
            break;
        case 0x3:
            Execute_Class0_Line3(self, fTable[param]);
            break;
        case 0x4: // LD@R-
            C[R] = param;
            DecrementPointer(&R);
            break;
        case 0x5:
            Execute_Class0_Line5(self, fTable[param]);
            break;
        case 0x6:
            Execute_Class0_Line6(self, param);
            break;
        case 0x7:
             Execute_Class0_Line7(self, fTable[param]);
            break;
        case 0x8:
            Execute_Class0_Line8(self, param);
            break;
        case 0xA: // WRITE r
            self->reg.ramreg = param;
            [self->bus writeRamLocation: param ofChip: self->reg.ramchip from: C];
            if (self->reg.peripheral) {
                //int i;
                //printf("Peripheral Write: %02X %X ", self->reg.peripheral, param);
                //for (i = 13; i >= 0; i--)
                //    printf("%X", C[i]);
                //printf("\n");
                [self->bus writeToRegister: param ofPeripheral: self->reg.peripheral from: C];
            }
            break;
        case 0xB:
            Execute_Class0_LineB(self, fTable[param]);
            break;
        case 0xC:
            Execute_Class0_LineC(self, param);
            break;
        case 0xE: // READ r or READ DATA
            if (param)
                self->reg.ramreg = param;
            if (self->reg.peripheral) {
                //printf("Peripheral Read: %02X %x \n", self->reg.peripheral, param);
                [self->bus readFromRegister: param ofPeripheral: self->reg.peripheral into: C];
            }
            else
                [self->bus readRamLocation: self->reg.ramreg ofChip: self->reg.ramchip into: C];
            break;
        case 0xF:
            Execute_Class0_LineF(self, fTable[param]);
            break;
        default: UNIMP;
    }
}

//---------------------------- Class 1 ---------------------------------------------

static void Execute_Class1(CPU *self, int instr) {
    //UNIMP;
    int word2;
    Fetch(self, &word2);
    switch (word2 & 0x3) {
        case 0: // NCX
            if (!self->reg.carry) goto execute; else return;
        case 1: // CX
            if (self->reg.carry) goto execute; else return;
        case 2: // NCG
            if (!self->reg.carry) goto jump; else return;
        case 3: // CG
            if (self->reg.carry) goto jump; else return;
    }
execute:
    PushReturnStack(self, PC);
jump:
    PC = ((word2 & 0x3fc) << 6) | ((instr & 0x3fc) >> 2);
}

//---------------------------- Class 2 ---------------------------------------------

static void Execute_Class2(CPU *self, int instr) {
    int opcode = instr >> 5;
    int field = (instr >> 2) & 0x7;
    int start, count;
    Bit carry, zero;
    static Digits_14 zeroes;
    Digits_14 scratch;
    switch (field) {
        case 0x0: // @R
            start = R;
            count = 1;
            break;
        case 0x1: // S&X
            start = 0;
            count = 3;
            break;
        case 0x2: // R<
            start = 0;
            count = R + 1;
            break;
        case 0x3: // ALL
            start = 0;
            count = 14;
            break;
        case 0x4: // P-Q
            start = P;
            count = (Q >= P) ? Q - P + 1 : 14 - P;
            break;
        case 0x5: // XS
            start = 2;
            count = 1;
            break;
        case 0x6: // M
            start = 3;
            count = 10;
            break;
        case 0x7: // MS
            start = 13;
            count = 1;
            break;
        default:
            UNIMP;
            return;
    }
    switch (opcode) {
        case 0x00: // A=0
            FillDigits(&A[start], 0, count);
            break;
        case 0x01: // B=0
            FillDigits(&B[start], 0, count);
            break;
        case 0x02: // C=0
            FillDigits(&C[start], 0, count);
            break;
        case 0x03: // A<>B
            ExchangeDigits(A, B, start, count);
            break;
        case 0x04: // B=A
            CopyDigits(&A[start], &B[start], count);
            break;
        case 0x05: // A<>C
            ExchangeDigits(A, C, start, count);
            break;
        case 0x06: // C=B
            CopyDigits(&B[start], &C[start], count);
            break;
        case 0x07: // C<>B
            ExchangeDigits(C, B, start, count);
            break;
        case 0x08: // A=C
            CopyDigits(&C[start], &A[start], count);
            break;
        case 0x09: // A=A+B
            carry = 0;
            AddOrSubtractDigits(ADD, MODE, A, B, A, start, count, &carry, &zero);
            self->nextCarry = carry;
            break;
        case 0x0A: // A=A+C
            carry = 0;
            AddOrSubtractDigits(ADD, MODE, A, C, A, start, count, &carry, &zero);
            self->nextCarry = carry;
            break;
        case 0x0B: // A=A+1
            carry = 1;
            AddOrSubtractDigits(ADD, MODE, A, zeroes, A, start, count, &carry, &zero);
            self->nextCarry = carry;
            break;
        case 0x0C: // A=A-B
            carry = 1;
            AddOrSubtractDigits(SUB, MODE, A, B, A, start, count, &carry, &zero);
            self->nextCarry = !carry;
            break;
        case 0x0D: // A=A-1
            carry = 0;
            AddOrSubtractDigits(SUB, MODE, A, zeroes, A, start, count, &carry, &zero);
            self->nextCarry = !carry;
            break;
        case 0x0E: // A=A-C
            carry = 1;
            AddOrSubtractDigits(SUB, MODE, A, C, A, start, count, &carry, &zero);
            self->nextCarry = !carry;
            break;
        case 0x0F: // C=C+C
            carry = 0;
            AddOrSubtractDigits(ADD, MODE, C, C, C, start, count, &carry, &zero);
            self->nextCarry = carry;
            break;
        case 0x10: // C=C+A
            carry = 0;
            AddOrSubtractDigits(ADD, MODE, C, A, C, start, count, &carry, &zero);
            self->nextCarry = carry;
            break;
        case 0x11: // C=C+1
            carry = 1;
            AddOrSubtractDigits(ADD, MODE, C, zeroes, C, start, count, &carry, &zero);
            self->nextCarry = carry;
            break;
        case 0x12: // C=A-C
            carry = 1;
            AddOrSubtractDigits(SUB, MODE, A, C, C, start, count, &carry, &zero);
            self->nextCarry = !carry;
            break;
        case 0x13: // C=C-1
            carry = 0;
            AddOrSubtractDigits(SUB, MODE, C, zeroes, C, start, count, &carry, &zero);
            self->nextCarry = !carry;
            break;
        case 0x14: // C=0-C
            carry = 1;
            AddOrSubtractDigits(SUB, MODE, zeroes, C, C, start, count, &carry, &zero);
            self->nextCarry = !carry;
            break;
        case 0x15: // C=-C-1
            carry = 0;
            AddOrSubtractDigits(SUB, MODE, zeroes, C, C, start, count, &carry, &zero);
            self->nextCarry = !carry;
        case 0x16: // ?B!=0
            carry = 1;
            AddOrSubtractDigits(SUB, MODE, B, zeroes, scratch, start, count, &carry, &zero);
            self->nextCarry = !zero;
            break;
        case 0x17: // ?C!=0
            carry = 1;
            AddOrSubtractDigits(SUB, MODE, C, zeroes, scratch, start, count, &carry, &zero);
            self->nextCarry = !zero;
            break;
        case 0x18: // ?A<C
            carry = 1;
            AddOrSubtractDigits(SUB, MODE, A, C, scratch, start, count, &carry, &zero);
            self->nextCarry = !carry;
            break;
        case 0x19: // ?A<B
            carry = 1;
            AddOrSubtractDigits(SUB, MODE, A, B, scratch, start, count, &carry, &zero);
            self->nextCarry = !carry;
            break;
        case 0x1A: // ?A!=0
            carry = 1;
            AddOrSubtractDigits(SUB, MODE, A, zeroes, scratch, start, count, &carry, &zero);
            self->nextCarry = !zero;
            break;
        case 0x1B: // ?A!=C
            carry = 1;
            AddOrSubtractDigits(SUB, MODE, A, C, scratch, start, count, &carry, &zero);
            self->nextCarry = !zero;
            break;
        case 0x1C: // RSHFA
            ShiftDigitsRight(A, start, count);
            break;
        case 0x1D: // RSHFB
            ShiftDigitsRight(B, start, count);
            break;
        case 0x1E: // RSHFC
            ShiftDigitsRight(C, start, count);
            break;
        case 0x1F: // LSHFA
            ShiftDigitsLeft(A, start, count);
            break;
        default:
            UNIMP;
    }
}

//---------------------------- Class 3 ---------------------------------------------

static void Execute_Class3(CPU *self, int instr) {
    int v = (instr >> 2) & 1;
    int disp = instr >> 3;
    if (disp >= 64)
        disp -= 128;
    if (!(v ^ self->reg.carry))
        PC = (PC - 1 + disp) & 0xffff;
}

//----------------------------------------------------------------------------------

static void ExecuteNextInstruction(CPU *self) {
    int instr;
    self->savedPC = PC;
    self->nextCarry = 0;
    if (Fetch(self, &instr)) {
        switch (instr & 0x3) {
            case 0:
                Execute_Class0(self, instr);
                break;
            case 1:
                Execute_Class1(self, instr);
                break;
            case 2:
                Execute_Class2(self, instr);
                break;
            case 3:
                Execute_Class3(self, instr);
                break;
        }
    }
    else
        PC = PopReturnStack(self);
    self->reg.carry = self->nextCarry;
    if (self->breakpointIsSet && PC == self->breakpointAddr)
        [self setRunning: NO];
}

//----------------------------------------------------------------------------------

@implementation CPU

- (void) reset {
    memset(&self->reg, 0, sizeof(self->reg));
    // The H41 firmware relies on the ON key being down on wakeup from
    // deep sleep, otherwise the display doesn't get turned on. But it does
    // a dummy test & reset of the keyboard before testing for this, so we
    // have to arrange for it to appear held down for at least 2 kbd tests.
    self->reg.KY = onKeyCode;
    self->keyReleaseDelay = 1;      
    [self setPowerMode: deepSleep];
    [self setPowerMode: powerOn];
}

- (void) setPowerMode: (enum PowerMode) mode {
    if (self->powerMode != mode) {
        if (mode == powerOn && self->powerMode == deepSleep)
            self->reg.carry = 1;
        //printf("Setting power mode to %d\n", mode);
        self->powerMode = mode;
        if (mode != powerOn)
            FlushAndSuspendSoundOutput();
        [self->debugController updateDisplays];
    }
}

- (BOOL) running {
    return self->powerMode == powerOn && self->runFlag;
}

- (void) setRunning: (BOOL) state {
    if (self->runFlag != state) {
        self->runFlag = state;
        [self->debugController updateButtons];
        [self->debugController updateDisplays];
    }
    self->simulationTime = [NSDate timeIntervalSinceReferenceDate];
}

- (void) step {
    ExecuteNextInstruction(self);
    [self->debugController updateDisplays];
}

- (void) timeSlice: (NSTimer *) timer {
    NSTimeInterval currentTime = [NSDate timeIntervalSinceReferenceDate];
    if ([self running]) {
        if (currentTime - self->simulationTime > maxSimulationTimeLag)
            self->simulationTime = currentTime - maxSimulationTimeLag;
        self->cycleLimit = SoundAvailableBufferSpace();
        while ([self running] && self->cycleLimit >= 2 && self->simulationTime < currentTime)
            ExecuteNextInstruction(self);
        [self->debugController updateDisplays];
    }
    else
        self->simulationTime = currentTime;
    SoundTimeslice();
}

//- (int) breakpoint {
//    // Return the currently set breakpoint, or -1 if none.
//    if (self->breakpointIsSet)
//        return self->breakpointAddr;
//    else
//        return -1;
//}

- (BOOL) getBreakpoint: (int *) addr {
    *addr = self->breakpointAddr;
    return self->breakpointIsSet;
}

- (void) setBreakpointAt: (Bits_16) addr {
    self->breakpointAddr = addr;
    self->breakpointIsSet = YES;
}

- (void) clearBreakpoint {
    self->breakpointIsSet = NO;
}

- (void) unimplementedInstruction {	
    [self abortInstruction: @"Unimplemented instruction"];
}

- (void) abortInstruction: (NSString *) message {
    PC = self->savedPC;
    [self->debugController show: YES];
    [self setRunning: NO];
    [self->debugController alert: message];
}

- (void) keyWithCode: (int) code pressed: (BOOL) pressed {
    //printf("CPU: KbdInterface: Key code 0x%02X %s\n",
    //    code, pressed ? "pressed" : "released");
    if (pressed) {
        int row = code >> 4;
        int col = code & 0x0f;
        self->reg.KY = row | keyColTable[col];
        self->reg.keyDown = YES;
        if (self->reg.KY == onKeyCode && self->powerMode != deepSleep) {
            //printf("Setting powerOffFlag\n");
            self->powerOffFlag = YES; // will enter deep sleep on next POWOFF
        }
        if (self->powerMode == lightSleep || self->reg.KY == onKeyCode)
            [self setPowerMode: powerOn];
    }
    else
        self->reg.keyDown = NO;
    [self->debugController updateDisplays];
}

@end
