//
//  CalculatorController.m
//  HP41
//
//  Created by greg on Wed Aug 13 2003.
//  Copyright (c) 2003 __MyCompanyName__. All rights reserved.
//

#import "CalculatorController.h"
#import "DebugController.h"
#import "DocViewerController.h"
#import "BuiltinMemory.h"
#import "RomChip.h"
#import "RamChip.h"
#import "Bus.h"
#import "CPU.h"
#import "Keyboard.h"
#import "Display.h"
#import "Documents.h"

static float timeSliceInterval = 0.05;

@implementation CalculatorController

- (id) init {
    self->docViewers = [[NSMutableDictionary dictionary] retain];
    return self;
}

- (void) dealloc {
    [self->docViewers release];
    [super dealloc];
}

- (void) installBuiltinRoms {
    int slot;
    NSString **p, *path;
    RomChip *rom;

    slot = 0;
    for (p = builtinRomRsrcNames; *p; p++) {
        if ([*p length] > 0) {
            path = [[NSBundle mainBundle] pathForResource: *p ofType: @"ROM"];
            rom = [RomChip fromFile: path];
            [self->bus installRomChip: rom inSlot: slot];
            slot++;
        }
    }
}

- (void) installBuiltinRam {
    RamChip *ram;
    struct RamDesc *ramdesc;
    int slot;
    
    for (ramdesc = builtinRamTable; ramdesc->firstChip >= 0; ramdesc++) {
        for (slot = ramdesc->firstChip; slot <= ramdesc->lastChip; slot++) {
            ram = [[RamChip alloc] initWithMask: ramdesc->registerMask];
            [self->bus installRamChip: ram inSlot: slot];
        }
    }
}

- (void) startExecutionTimer {
    self->executionTimer = [NSTimer scheduledTimerWithTimeInterval: timeSliceInterval 
        target: self selector: @selector(timeSlice:) userInfo: nil repeats: YES];
    [self->cpu setPowerMode: powerOn];
}

- (NSData *) getMemoryContents {
    int ramSize = 0x400;
    int addr;
    
    NSMutableData *data = [NSMutableData dataWithLength: 14 * ramSize];
    Digit *dp = [data mutableBytes];
    for (addr = 0; addr < ramSize; addr++) {
        [self->bus readRamLocation: addr & 0xf ofChip: addr >> 4 into: dp];
        dp += 14;
    }
    return data;
}

- (void) setMemoryContents: (NSData *) data {
    int ramSize = 0x400;
    int dataSize = [data length] / 14;
    int addr;
    
    if (dataSize > ramSize)
        dataSize = ramSize;
    const Digit *dp = [data bytes];
    for (addr = 0; addr < dataSize; addr++) {
        [self->bus writeRamLocation: addr & 0xf ofChip: addr >> 4 from: dp];
        dp += 14;
    }
}

- (void) saveMemory {
    NSData *data = [self getMemoryContents];
    [[NSUserDefaults standardUserDefaults] setObject: data forKey: @"memory"];
}

- (void) restoreMemory {
    NSData *data = [[NSUserDefaults standardUserDefaults] objectForKey: @"memory"];
    if (data)
        [self setMemoryContents: data];
}

- (void) awakeFromNib {
    BOOL x = [self->calculatorWindow isExcludedFromWindowsMenu];
    x = x;
    [self->calculatorWindow setExcludedFromWindowsMenu: NO];
    [self installBuiltinRoms];
    [self installBuiltinRam];
    [self restoreMemory];
    [self startExecutionTimer];
    [self->cpu reset];
    [self->cpu setRunning: YES];
}

- (void) applicationWillTerminate: (id) notification {
    [self saveMemory];
}

- (void) timeSlice: (NSTimer *) timer {
    [self->cpu timeSlice: timer];
    [self->display timeSlice: timer];
}

- (void) showDoc: (DocumentDesc *) desc {
    DocViewerController *docViewer = [self->docViewers objectForKey: desc->rsrcName];
    if (!docViewer) {
        docViewer = [[DocViewerController alloc] initWithDesc: desc];
        if (!docViewer)
            return;
        [self->docViewers setObject: docViewer forKey: desc->rsrcName];
    }
    [docViewer show: YES];
}

- (void) hideDoc: (DocumentDesc *) desc {
    [[self->docViewers objectForKey: desc->rsrcName] show: FALSE];
}

- (BOOL) docIsShown: (DocumentDesc *) desc {
    return [[self->docViewers objectForKey: desc->rsrcName] isShown];
}

- (void) showOrHideDoc: (DocumentDesc *) desc {
    if (![self docIsShown: desc])
        [self showDoc: desc];
    else
        [self hideDoc: desc];
}

- (BOOL) validateMenuItem: (id) item {
    SEL action = [item action];
    if (action == @selector(showRegistersCmd:)) {
        [item setState: [self->registerWindow isVisible]];
        return YES;
    }
    if (action == @selector(showCalculatorManualCmd:)) {
        [item setState: [self docIsShown: ownersManual]];
        return YES;
    }
    return NO;
}

- (IBAction) showRegistersCmd: (id) item {
    [self->debugController show: ![self->registerWindow isVisible]];
}

- (IBAction) showCalculatorManualCmd: (id) item {
    [self showOrHideDoc: ownersManual];
}

- (void) keyDown: (NSEvent *) event {
    [self->keyboard keyEvent: event down: YES];
}

- (void) keyUp: (NSEvent *) event {
    [self->keyboard keyEvent: event down: NO];
}

@end
