//
//  DebugController.m
//  HP41
//
//  Created by greg on Wed Aug 13 2003.
//  Copyright (c) 2003 __MyCompanyName__. All rights reserved.
//

#import "DebugController.h"
#import "DigitsDisplay.h"
#import "BitsDisplay.h"
#import "ModeDisplay.h"
#import "RamDisplay.h"
#import "RomDisplay.h"
#import "CPURegisters.h"
#import "CPU.h"
#import "Bus.h"
#import "RomChip.h"
#import "Display.h"
#import "Disassembly.h"

enum {
    epromAddress = 0x4000
};

@implementation DebugController

- (void) awakeFromNib {
    CPURegisters *reg = &self->cpu->reg;
    DisplayRegisters *dreg = &self->display->registers;
    [self->displayA      attachTo: reg->A         digits: 14];
    [self->displayB      attachTo: reg->B         digits: 14];
    [self->displayC      attachTo: reg->C         digits: 14];
    [self->displayM      attachTo: reg->M         digits: 14];
    [self->displayN      attachTo: reg->N         digits: 14];
    [self->displayP      attachTo: &reg->PQ[0]    digits: 1];
    [self->displayQ      attachTo: &reg->PQ[1]    digits: 1];
    [self->displayPC     attachTo: &reg->PC       bits: 16];
    [self->displayG      attachTo: &reg->G        bits: 8];
    [self->displayST     attachTo: &reg->ST       bits: 8];
    [self->displayT      attachTo: &reg->T        bits: 8];
    [self->displayFI     attachTo: &reg->FI       bits: 14];
    [self->displayKY     attachTo: &reg->KY       bits: 8];
    [self->displayXST    attachTo: &reg->XST      bits: 6];
    [self->displayStack0 attachTo: &reg->stack[0] bits: 16];
    [self->displayStack1 attachTo: &reg->stack[1] bits: 16];
    [self->displayStack2 attachTo: &reg->stack[2] bits: 16];
    [self->displayStack3 attachTo: &reg->stack[3] bits: 16];
    [self->displayRselect       attachTo: &reg->r          withValues: "PQ"];
    [self->displayCarry         attachTo: &reg->carry      bits: 1];
    [self->displayMode          attachTo: &reg->mode       withValues: "DH"];
    [self->displayRamChipSelect attachTo: &reg->ramchip    bits: 6];
    [self->displayRamRegSelect  attachTo: &reg->ramreg     bits: 4];
    [self->displayPeriphSelect  attachTo: &reg->peripheral bits: 8];
    [self->displayKeyDown       attachTo: &reg->keyDown    bits: 1];
    [self->displayPowerMode     attachTo: &self->cpu->powerMode withValues: "DLP"];
    [self->displayDispA         attachTo: dreg->A          digits: 12];
    [self->displayDispB         attachTo: dreg->B          digits: 12];
    [self->displayDispC         attachTo: dreg->C          digits: 12];
    [self->displayDispE         attachTo: &dreg->E         bits: 12];
    [self->displayCurrentRam    attachTo: &reg->ramchip];
    [self updateButtons];
    RomChip *rom = [RomChip blankEprom];
    [rom writeLocation: 0 from: 0x3E0]; // RTN
    [self->bus installRomChip: rom inSlot: epromAddress >> 12];
}

- (void) show: (BOOL) shown {
    if (shown)
        [self->registerWindow makeKeyAndOrderFront: self];
    else
        [self->registerWindow close];
}

- (BOOL) running {
    return [self->cpu running];
}

- (int) pc; {
    return self->cpu->reg.PC;
}

//- (int) breakpoint {
//    return [self->cpu breakpoint];
//}

- (BOOL) getBreakpoint: (int *) addr {
    return [self->cpu getBreakpoint: addr];
}

- (BOOL) readRomLocation: (int) addr into: (int *) data {
    return [self->bus readRomLocation: addr into: data];
}

- (void) readRamLocation: (int) reg ofChip: (int) slot into: (Digits_14) data {
    [self->bus readRamLocation: reg ofChip: slot into: data];
}

- (IBAction) halt: (id) sender {
    [self->cpu setRunning: NO];
}

- (IBAction) step: (id) sender {
    [self->cpu step];
}

- (IBAction) run: (id) sender {
    [self->cpu setRunning: YES];
}

- (IBAction) incPc: (id) sender {
    ++self->cpu->reg.PC;
    self->cpu->reg.PC &= 0xffff;
    [self updateDisplays];
}

- (IBAction) decPc: (id) sender {
    --self->cpu->reg.PC;
    self->cpu->reg.PC &= 0xffff;
    [self updateDisplays];
}

- (IBAction) reset: (id) sender {
    [self->cpu reset];
    [self updateDisplays];
}

- (IBAction) resetToEprom: (id) sender {
    [self->cpu reset];
    self->cpu->reg.PC = epromAddress + 1;
    [self->cpu setPowerMode: powerOn];
    [self updateDisplays];
}

- (IBAction) programEprom: (id) sender {
    NSString *text = [[self->epromField textStorage] string];
    NSScanner *scanner = [NSScanner scannerWithString: text];
    unsigned int addr = 0x8000;
    unsigned int word;
    while ([scanner scanHexInt: &word]) {
        word &= 0xffff;
        if ([scanner scanString: @":" intoString: nil])
            addr = word & 0xffff;
        else {
            if (![self->bus writeRomLocation: addr from: word & 0x3ff]) {
                NSBeginAlertSheet(@"Unable to write to ROM", @"OK", nil, nil,
                    self->registerWindow, nil, NULL, NULL, nil,
                    @"Location %04X is not writable.", addr);
                break;
            }
            ++addr;
        }
    }
    [self->displayRom setNeedsDisplay: YES];
}


- (void) updateButtons {
    BOOL running = self->cpu->runFlag;
    [self->haltButton setEnabled: running];
    [self->stepButton setEnabled: !running];
    [self->runButton setEnabled: !running];
    [self->incPcButton setEnabled: !running];
    [self->decPcButton setEnabled: !running];
}

- (void) updateDisplays {
    [[self->registerWindow contentView] setNeedsDisplay: YES];
}

- (void) registerChanged: (RegisterDisplay *) regDisplay {
    if (regDisplay == self->displayPC)
        [self->displayRom setNeedsDisplay: YES];
    else if (regDisplay == displayRamChipSelect)
        [self->displayCurrentRam setNeedsDisplay: YES];
}

- (int) numberMode {
    return [self->numberBaseSelector selectedColumn];
}

- (NSString *) formatAddress: (int) addr {
    return FormatRomAddressUsingMode([self numberMode], addr);
}

- (int) parseAddress: (NSString *) text {
    int addr, page;
    NSString *digits;
    NSScanner *scanner = [NSScanner scannerWithString: text];
    NSCharacterSet *digset = [NSCharacterSet 
        characterSetWithCharactersInString: @"0123456789ABCDEFabcdef"];
    [scanner scanCharactersFromSet: digset intoString: &digits];
    switch ([self numberMode]) {
        case disasmHex:
            addr = strtol([digits UTF8String], NULL, 16);
            break;
        case disasmOctal:
            addr = strtol([digits UTF8String], NULL, 8);
            if ([scanner scanString: @"," intoString: nil]) {
                page = addr;
                [scanner scanCharactersFromSet: digset intoString: &digits];
                addr = (page << 9) + strtol([digits UTF8String], NULL, 8);
            }
            break;
    }
    return addr;
}    

- (IBAction) updateBreakpoint: (id) sender {
    if ([self->breakpointSwitch state] == NSOnState) {
        int addr = [self parseAddress: [self->breakpointField stringValue]];
        [self->cpu setBreakpointAt: addr];
    }
    else
        [self->cpu clearBreakpoint];
    [self->displayRom setNeedsDisplay: YES];
}

- (void) updateBreakpointField {
    int addr;
    [self->cpu getBreakpoint: &addr];
    NSString *text = [self formatAddress: addr];
    [self->breakpointField setStringValue: text];
}

- (IBAction) numberModeControlChanged: (id) sender {
    [self updateBreakpointField];
    [self->displayRom numberModeControlChanged: sender];
}

- (IBAction) gotoAddress: (id) sender {
    int addr = [self parseAddress: [self->gotoField stringValue]];
    self->cpu->reg.PC = addr;
    [self updateDisplays];
}

//- (void) reportUnimplementedInstruction {
//    [self alert: @"Unimplemented instruction"];
//}

- (void) alert: (NSString *) message {
    [self->cpu setRunning: NO];
    NSBeginAlertSheet(message, @"OK", nil, nil,
    	self->registerWindow, nil, NULL, NULL, nil, @"");
}

@end
