//
//  Display.m
//  HP41
//
//  Created by greg on Sun Aug 10 2003.
//  Copyright (c) 2003 __MyCompanyName__. All rights reserved.
//

#import "Display.h"
#import "Bus.h";
#import "DebugController.h"
#import "DataOperations.h"

static NSString *segmentPathsRsrcName = @"hpchar";
static NSString *displayFontRsrcName = @"hpfont";
static NSString *annunciatorFontName = @"Helvetica";
static float annunciatorFontSize = 9;
static float annunciatorFontScale = 2;
static float annunciatorBottomMargin = 2.0;

DisplaySegmentMap punctSegmentTable[5] = {
    0x00000, // no punctuation
    0x08000, // .
    0x0C000, // :
    0x18000, // ,
    0x1C000  // ;  (only used during startup segment test)
};

typedef enum {LEFT, RIGHT} DispShiftDir;
typedef enum {SHORT, LONG} DispTransSize;
typedef enum {RA = 1, RB = 2, RC = 4, 
              RAB = RA + RB, RABC = RA + RB + RC} DispRegSet;

static void RotateRegisterLeft(Digits_12 reg) {
    int i;
    int d = reg[11];
    for (i = 11; i >= 1; i--)
        reg[i] = reg[i-1];
    reg[0] = d;
}

static void RotateRegisterRight(Digits_12 reg) {
    int i;
    int d = reg[0];
    for (i = 0; i <= 10; i++)
        reg[i] = reg[i+1];
    reg[11] = d;
}

static void FetchDigit(DispShiftDir dir, Digits_12 reg, Digit *dst) {
    // Fetch a digit from the appropriate end of the given register
    // into the specified destination, and rotate the register in
    // the specified direction.
    switch (dir) {
        case LEFT:
            *dst = reg[11];
            RotateRegisterLeft(reg);
            break;
        case RIGHT:
            *dst = reg[0];
            RotateRegisterRight(reg);
            break;
    }
}

static void ShiftDigit(DispShiftDir dir, Digits_12 reg, Digit src) {
    // Rotate the given digit into the specified register in the 
    // specified direction.
    switch (dir) {
        case LEFT:
            RotateRegisterLeft(reg);
            reg[0] = src;
            break;
        case RIGHT:
            RotateRegisterRight(reg);
            reg[11] = src;
            break;
    }
}

static void Fetch(DisplayRegisters *regs, 
    DispShiftDir dir, DispTransSize size, DispRegSet regset,
    Digits_14 data)
{
    // Fetch digits from the specified registers, rotating them in the
    // specified direction, and assemble them into the specified
    // destination. For size == LONG, fetches a total of 12 digits 
    // into the destination; for size == SHORT, fetches one digit
    // from each specified register.
    int cp = 0;
    while (cp < 12) {
        if (regset & RA) FetchDigit(dir, regs->A, &data[cp++]);
        if (regset & RB) FetchDigit(dir, regs->B, &data[cp++]);
        if (regset & RC) FetchDigit(dir, regs->C, &data[cp++]);
        if (size == SHORT)
            break;
    }
}

static void Shift(DisplayRegisters *regs, 
    DispShiftDir dir, DispTransSize size, DispRegSet regset,
    Digits_14 data)
{
    // Distribute digits from the given source and rotate them
    // into the specified registers. For size == LONG, shifts a 
    // total of 12 digits from the source; for size == SHORT,
    // shifts one digit into each specified register.
    int cp = 0;
    while (cp < 12) {
        if (regset & RA) ShiftDigit(dir, regs->A, data[cp++]);
        if (regset & RB) ShiftDigit(dir, regs->B, data[cp++]);
        if (regset & RC) ShiftDigit(dir, regs->C, data[cp++] & 0x1);
        if (size == SHORT)
            break;
    }
}

static NSString *annunciatorStrings[numAnnunciators] = {
    @"BAT  ",
    @"USER  ",
    @"G",
    @"RAD  ",
    @"SHIFT  ",
    @"0",
    @"1",
    @"2",
    @"3",
    @"4  ",
    @"PRGM  ",
    @"ALPHA",
};

static void CalculateAnnunciatorPositions(NSPoint positions[], NSFont *font, NSRect bounds) {
    // Distribute the annunciators evenly across the width of the display
    // based on the sizes of their strings.
    NSLayoutManager *layoutManager = [[NSLayoutManager alloc] init];
    int i;
    float annunciatorWidths[numAnnunciators];
    float spaceWidth, x, y, d, h;
    float totalWidth = 0.0;
    for (i = 0; i < numAnnunciators; i++) {
        float width = [annunciatorStrings[i] sizeWithAttributes:nil].width;
        annunciatorWidths[i] = width;
        totalWidth += width;
    }
    spaceWidth = (bounds.size.width - totalWidth) / (numAnnunciators - 1);
    d = -[font descender];
    h = [layoutManager defaultLineHeightForFont:font];
    y = bounds.size.height - annunciatorBottomMargin - (h - d) / annunciatorFontScale;
    //y = 15;
    x = 0.0;
    for (i = 0; i < numAnnunciators; i++) {
        positions[i] = NSMakePoint(x, y);
        x += annunciatorWidths[i] + spaceWidth;
    }
}

@implementation Display

- (BOOL) isFlipped {
    return YES;
}

- (void) awakeFromNib {
    //int i;
    self->foregroundColor = [[NSColorList colorListNamed: @"HP41"] colorWithKey: @"displayForegroundColor"];
    LoadFont(self->displayFont, displayFontRsrcName);
    LoadSegmentPaths(self->segmentPaths, segmentPathsRsrcName);
    self->annunciatorFont = [NSFont fontWithName: annunciatorFontName size: annunciatorFontScale * annunciatorFontSize];
    CalculateAnnunciatorPositions(self->annunciatorPositions, self->annunciatorFont, [self bounds]);
    self->on = YES;
    [self->bus installPeripheral: self inSlot: 0xFD];
    if (1) {
        int i;
        for (i = 0; i < numDisplayCells; i++) {
            self->registers.A[i] = 0xA;
            self->registers.B[i] = 0x3;
            self->registers.C[i] = 0x2;
        }
        self->registers.E = 0xfff;
    }
}

// We arrange for a slight delay between changing the display registers and
// updating the screen, so that intermediate display states which would
// normally last for an imperceptibly short time don't get flashed if we
// happen to get a draw event in between.

- (void) scheduleUpdate {
    self->updateCountdown = 2;
}

- (void) timeSlice: (NSTimer *) timer {
    if (self->updateCountdown)
        if (--self->updateCountdown == 0)
            [self setNeedsDisplay: YES];
}

- (void) displayOff {
    // Turn the display off.
    if (self->on) {
        self->on = NO;
        [self scheduleUpdate];
    }
}

- (void) displayToggle {
    // Toggle the display between on and off.
    self->on = !self->on;
    [self scheduleUpdate];
}

- (void) pluggedIntoBus: (Bus *) b {
    self->bus = b;
}

- (void) readFromRegister: (Bits_4) reg into: (Digits_14) data {
    // Implement READ f or READ DATA instruction with display as selected peripheral.
    DisplayRegisters *regs = &self->registers;
    switch (reg) {
        case 0x2: /*FLLDC  */   Fetch(regs, LEFT, LONG, RC, data); break;
        case 0x3: /*FLLDAB */   Fetch(regs, LEFT, LONG, RAB, data); break;   // Frederic
        case 0x4: /*FLLABC*/    Fetch(regs, LEFT, LONG, RABC, data); break;    // Frederic
        case 0x5: /*READDEN*/   BitsToDigits(self->registers.E, data, 4); return; // doesn't change display
        case 0xE: /*FRSABC */   Fetch(regs, RIGHT, SHORT, RABC, data); break;
        case 0xF: /*FLSABC */   Fetch(regs, LEFT, SHORT, RABC, data); break;
        default:
            [self->bus abortInstruction: @"Unimplemented display operation"];
    }
    [self scheduleUpdate];
}

- (void) writeToRegister: (Bits_4) reg from: (Digits_14) data {
    // Implement WRITE f instruction with display as selected peripheral.
    DisplayRegisters *regs = &self->registers;
    switch (reg) {
        case 0x0: /*SRLDA */ Shift(regs, RIGHT, LONG, RA, data); break;
        case 0x1: /*SRLDB */ Shift(regs, RIGHT, LONG, RB, data); break;
        case 0x2: /*SRLDC */ Shift(regs, RIGHT, LONG, RC, data); break;
        case 0x3: /*SRLDAB*/ Shift(regs, RIGHT, LONG, RAB, data); break;  // Frederic
        case 0x4: /*SRLABC*/ Shift(regs, RIGHT, LONG, RABC, data); break;
        case 0xE: /*SRSABC*/ Shift(regs, RIGHT, SHORT, RABC, data); break;
        case 0xF: /*SLSABC*/ Shift(regs, LEFT, SHORT, RABC, data); break;
        default:
            [self->bus abortInstruction: @"Unimplemented display operation"];
    }
    [self scheduleUpdate];
}

- (void) writeDataFrom: (Digits_14) data {
    // Implement WRITE DATA instruction with display as selected peripheral.
    DigitsToBits(data, &self->registers.E, 12);
    [self scheduleUpdate];
}

- (DisplaySegmentMap) segmentsForCell: (int) i {
    // Determine which segments should be on for cell i based on
    // the contents of the display registers. Note that cells are
    // numbered from left to right, which is the opposite of the
    // digit numbering in the display registers.
    DisplayRegisters *r = &self->registers;
    int j = 11 - i;
    int nineBitCode = ((r->C[j] << 8) | (r->B[j] << 4) | (r->A[j])) & 0x1ff;
    int charCode = ((nineBitCode & 0x100) >> 2) | (nineBitCode & 0x3f);
    int punctCode = ((r->C[j] & 0x2) << 1) | (nineBitCode & 0xc0) >> 6;
    return self->displayFont[charCode] | punctSegmentTable[punctCode];
}

- (BOOL) annunciatorOn: (int) i {
    // Determine whether annunciator number i should be on based
    // on the contents of the E register. Note that i is not the
    // same as the bit number in E, because the annunciators are
    // numbered from left to right.
    int j = 11 - i;
    return (self->registers.E & (1 << j)) != 0;
}

- (void) drawRect: (NSRect) rect {
    if (self->on) {
        int i, seg, segmentsOn;
        id attrs;
        id cellTranslation = [NSAffineTransform transform];
        [cellTranslation translateXBy: [self cellWidth] yBy: 0.0];
        [self->foregroundColor set];
        if (1) {
            [NSGraphicsContext saveGraphicsState];
            for (i = 0; i < numDisplayCells; i++) {
                //segmentsOn = self->cells[i];
                segmentsOn = [self segmentsForCell: i];
                for (seg = 0; seg < numDisplaySegments; seg++) {
                    if (segmentsOn & (1 << seg))
                        [self->segmentPaths[seg] fill];
                }
                [cellTranslation concat];
            }
            [NSGraphicsContext restoreGraphicsState];
        }
        attrs = [NSDictionary dictionaryWithObject: self->annunciatorFont forKey: NSFontAttributeName];
        for (i = 0; i < numAnnunciators; i++)
            //if (self->annunciators[i]) {
            if ([self annunciatorOn: i]) {
                id t = [NSAffineTransform transform];
                NSPoint p = self->annunciatorPositions[i];
                [t translateXBy: p.x yBy: p.y];
                [t scaleBy: 1 / annunciatorFontScale];
                [NSGraphicsContext saveGraphicsState];
                [t concat];
                [annunciatorStrings[i] drawAtPoint: NSMakePoint(0, 0) withAttributes: attrs];
                [NSGraphicsContext restoreGraphicsState];
        }
    }
}

- (float) cellWidth {
   return [self bounds].size.width / numDisplayCells;
}

- (BOOL) acceptsFirstMouse: (NSEvent *) event {
    return YES;
}

@end
