#import "CPU_KbdInterface.h"
#import "Key.h"
#import "KeyGroup.h"
#import "Rocker.h"
#import "Keyboard.h"

// Keys on the keyboard are divided into groups having a
// regular row and column spacing. The following tables are
// used to map from the position of a key within its group
// to a key code. The first and second hex digits of the
// key code correspond to its row and column number in the
// keyboard scanning matrix, respectively.

static struct KeyMap functionMap = {
    4, 5, {
        {0x00, 0x01, 0x02, 0x03, 0x04},
        {0x10, 0x11, 0x12, 0x13, 0x14},
        {0x20, 0x21, 0x22, 0x23, 0x24},
        {0x30, 0x00, 0x32, 0x33, 0x34}
    }
};

static struct KeyMap arithmeticMap = {
    4, 1, {
        {0x40},
        {0x50},
        {0x60},
        {0x70}
    }
};

static struct KeyMap numberMap = {
    4, 3, {
        {0x41, 0x42, 0x43},
        {0x51, 0x52, 0x53},
        {0x61, 0x62, 0x63},
        {0x71, 0x72, 0x73}
    }
};

// Rocker switches are described by a RockerDescriptor structure
// which gives the names of the images for the two pressed-down
// states and the corresponding key codes.

static struct RockerDescriptor on_user_rocker_descriptor = {
    {@"on.tiff", @"user.tiff"},
    {0x80, 0x64}
};

static struct RockerDescriptor prgm_alpha_rocker_descriptor = {
    {@"prgm.tiff", @"alpha.tiff"},
    {0x54, 0x44}
};

// The following table maps Macintosh keyboard characters to
// key codes.

struct MacKeyboardMap {
    int c;
    int code;
};

#define Control(c) c & 0x1f
#define C_O Control('O')
#define C_U Control('U')
#define C_P Control('P')
#define C_A Control('A')
#define C_R Control('R')
#define C_S Control('S')
#define ESC 27
#define ENT 0x03 // ENTER key on Mac keyboard
#define RET 0x0d
#define DEL 0x7f

static struct MacKeyboardMap macKeyboardTable[] = {
    {C_O, 0x80}, {C_U, 0x64},              {C_P, 0x54}, {C_A, 0x44},
    
    {'A', 0x00}, {'B', 0x01}, {'C', 0x02}, {'D', 0x03}, {'E', 0x04},
    {'F', 0x10}, {'G', 0x11}, {'H', 0x12}, {'I', 0x13}, {'J', 0x14},
    {ESC, 0x20}, {'K', 0x21}, {'L', 0x22}, {'M', 0x23}, {C_S, 0x24},
    {'N', 0x30},              {'O', 0x32}, {'P', 0x33}, {DEL, 0x34},
    {RET, 0x30},
    {ENT, 0x30},

    {'Q', 0x40},             {'R', 0x41}, {'S', 0x42}, {'T', 0x43},
    {'U', 0x50},             {'V', 0x51}, {'W', 0x52}, {'X', 0x53},
    {'Y', 0x60},             {'Z', 0x61}, {'=', 0x62}, {'?', 0x63},
    {':', 0x70},             {' ', 0x71}, {',', 0x72}, {C_R, 0x73},
    
    {'-', 0x40},             {'7', 0x41}, {'8', 0x42}, {'9', 0x43},
    {'+', 0x50},             {'4', 0x51}, {'5', 0x52}, {'6', 0x53},
    {'*', 0x60},             {'1', 0x61}, {'2', 0x62}, {'3', 0x63},
    {'/', 0x70},             {'0', 0x71}, {'.', 0x72},    

    {0, 0}
};

// The following table maps key codes to Keys and Rockers, so that visual feedback
// can be provided when a key is activated by a Macintosh key event.

static id keyCodeToKeyTable[256];

@implementation Keyboard

// Install the Rockers' images and key codes. 

- (void) awakeFromNib {
    [self->on_user_rocker setUpFromDescriptor: &on_user_rocker_descriptor];
    [self->prgm_alpha_rocker setUpFromDescriptor: &prgm_alpha_rocker_descriptor];    
}

// Key groups call the following method during initialisation to
// find their key maps.

- (struct KeyMap *) keyMapForGroup: (KeyGroup *) group {
    if (group == self->arithmeticGroup)
        return &arithmeticMap;
    if (group == self->functionGroup)
        return &functionMap;
    if (group == self->numberGroup)
        return &numberMap;
    return NULL; // not reached
}

// KeyGroups and Rockers call the following method during intialisation
// to install a Key or Rocker in the keycode-to-key mapping.

- (void) registerKey: (id) key withCode: (int) code {
    keyCodeToKeyTable[code] = key;
}

// The KeyGroups and Rockers send keyWithCode:pressed: to the Keyboard when
// a key is pressed or released. We pass it on to the CPU.

- (void) keyWithCode: (int) code pressed: (BOOL) pressed {
    [self->cpu keyWithCode: code pressed: pressed];
}

// Macintosh keyboard event handling

- (BOOL) acceptsFirstResponder {
    return YES;
}

- (BOOL) needsPanelToBecomeKey {
    return YES;
}

- (BOOL) becomeFirstResponder {
    return YES;
}

- (void) keyEvent: (NSEvent *) event down: (BOOL) down {
    NSString *chars = [[event characters] uppercaseString];
    int c = [chars characterAtIndex: 0];
    //printf("Keyboard keyEvent: 0x%04X\n", c);
    struct MacKeyboardMap *map = macKeyboardTable;
    while (map->c) {
        if (map->c == c) {
            [keyCodeToKeyTable[map->code] showFeedback: down forCode: map->code];
            [self keyWithCode: map->code pressed: down];
            return;
        }
        ++map;
    }
}

- (void) keyDown: (NSEvent *) event {
    [self keyEvent: event down: YES];
}

- (void) keyUp: (NSEvent *) event {
    [self keyEvent: event down: NO];
}

@end
