#import "Keyboard.h"
#import "Rocker.h"

@interface Rocker (private)

- (int) keyCode;

@end

@implementation Rocker

// The following method is called by the Keyboard during initialisation
// to install the rocker's images and key codes.

- (void) setUpFromDescriptor: (struct RockerDescriptor *) desc {
    self->images[0] = [NSImage imageNamed: desc->imageNames[0]];
    self->images[1] = [NSImage imageNamed: desc->imageNames[1]];
    self->keyCodes = desc->keyCodes;
    [self->keyboard registerKey: self withCode: self->keyCodes[0]];
    [self->keyboard registerKey: self withCode: self->keyCodes[1]];
}

// Change rocker state and redisplay.

- (void) setState: (enum RockerState) newState {
    self->state = newState;
    [self setNeedsDisplay: YES];
}

// Provide feedback for when activated by a Macintosh key event.

- (void) showFeedback: (BOOL) down forCode: (int) code {
    if (down) {
        if (code == self->keyCodes[0])
            [self setState: rocker_left];
        else
            [self setState: rocker_right];
    }
    else
        [self setState: rocker_centre];
}

// To draw a pressed-down rocker, we first cover up the non-pressed
// background image of the rocker with black, and then overlay the
// appropriate pressed-down image. If the rocker is not pressed,
// we do nothing and let the background image show through.

- (void) drawRect: (NSRect) rect
{
    if (self->state != rocker_centre) {
        NSSize size = [self bounds].size;
        int i = self->state - rocker_left;
        NSRect blackRect = NSMakeRect(0, 2, size.width, size.height - 4);
        [[NSColor blackColor] set];
        NSRectFill(blackRect);
        [self->images[i] compositeToPoint: NSMakePoint(0, 0) 
            operation: NSCompositeSourceAtop];
    }
}

// When the mouse is pressed or released, we update our state
// accordingly and send a message to the Keyboard.

- (void) mouseDown: (NSEvent *) event {
    NSPoint p = [self convertPoint: [event locationInWindow] fromView: nil];
    int width = [self bounds].size.width;
    if (p.x < width / 2)
        self->state = rocker_left;
    else
        self->state = rocker_right;
    [self setNeedsDisplay: YES];
    [self->keyboard keyWithCode: [self keyCode] pressed: YES];
}

- (void) mouseUp: (NSEvent *) event {
    [self->keyboard keyWithCode: [self keyCode] pressed: NO];
    self->state = rocker_centre;
    [self setNeedsDisplay: YES];
}

// The following method extracts the key code for the current
// state from the rocker's descriptor. (It is only called when
// the rocker is pressed.)

- (int) keyCode {
    return self->keyCodes[self->state - rocker_left];
}

// We need to intercept mouseDragged events too, otherwise they
// pass through to the ShapedWindow and it starts getting dragged
// around the screen.

- (void) mouseDragged: (NSEvent *) e {
}

- (BOOL) acceptsFirstMouse: (NSEvent *) event {
    return YES;
}

@end
